#
# Copyright © 2020 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.
"""

"""

import numpy as np
import pytest

import pyronn_torch


def test_init():
    assert pyronn_torch.cpp_extension


@pytest.mark.parametrize('with_texture', ('with_texture', False))
@pytest.mark.parametrize('with_backward', ('with_backward', False))
def test_projection(with_texture, with_backward):
    projector = pyronn_torch.ConeBeamProjector(
        (128, 128, 128), (2.0, 2.0, 2.0), (-127.5, -127.5, -127.5),
        (2, 480, 620), [1.0, 1.0], (0, 0),
        np.array(
            [[[-3.10e+2, -1.20e+03, 0.00e+00, 1.86e+5],
              [-2.40e+2, 0.00e+00, 1.20e+03, 1.44e+5],
              [-1.00e+00, 0.00e+00, 0.00e+00, 6.00e+2]],
             [[-2.89009888e+2, -1.20522754e+3, -1.02473585e-13, 1.86000000e+5],
              [-2.39963440e+2, -4.18857765e+0, 1.20000000e+3, 1.44000000e+5],
              [-9.99847710e-01, -1.74524058e-2, 0.00000000e+0,
               6.00000000e+2]]]))

    volume = projector.new_volume_tensor(
        requires_grad=True if with_backward else False)

    volume += 1.
    result = projector.project_forward(volume, use_texture=with_texture)

    assert result is not None
    if with_backward:
        assert volume.requires_grad
        assert result.requires_grad

        loss = result.mean()
        loss.backward()


@pytest.mark.parametrize('with_texture', ('with_texture', False))
@pytest.mark.parametrize('with_backward', ('with_backward', False))
def test_projection_backward(with_texture, with_backward):
    projector = pyronn_torch.ConeBeamProjector(
        (128, 128, 128), (2.0, 2.0, 2.0), (-127.5, -127.5, -127.5),
        (2, 480, 620), [1.0, 1.0], (0, 0),
        np.array(
            [[[-3.10e+2, -1.20e+03, 0.00e+00, 1.86e+5],
              [-2.40e+2, 0.00e+00, 1.20e+03, 1.44e+5],
              [-1.00e+00, 0.00e+00, 0.00e+00, 6.00e+2]],
             [[-2.89009888e+2, -1.20522754e+3, -1.02473585e-13, 1.86000000e+5],
              [-2.39963440e+2, -4.18857765e+0, 1.20000000e+3, 1.44000000e+5],
              [-9.99847710e-01, -1.74524058e-2, 0.00000000e+0,
               6.00000000e+2]]]))
    projection = projector.new_projection_tensor(
        requires_grad=True if with_backward else False)

    projection += 1.

    result = projector.project_backward(projection, use_texture=with_texture)
    assert result.shape == projector._volume_shape

    assert result is not None
    if with_backward:
        assert projection.requires_grad
        assert result.requires_grad

        loss = result.mean()
        loss.backward()


@pytest.mark.parametrize('with_backward', ('with_backward', False))
def test_conrad_config(with_backward, with_texture=True):
    import pytest
    pytest.importorskip("pyconrad")

    projector = pyronn_torch.ConeBeamProjector.from_conrad_config()

    volume = projector.new_volume_tensor(
        requires_grad=True if with_backward else False)

    volume += 1.
    result = projector.project_forward(volume, use_texture=with_texture)
    import pyconrad.autoinit
    pyconrad.imshow(result)

    assert result is not None
    if with_backward:
        assert volume.requires_grad
        assert result.requires_grad

        loss = result.mean()
        loss.backward()


def test_projection_backward_conrad(with_texture=True, with_backward=True):
    import pytest
    pytest.importorskip("pyconrad")

    projector = pyronn_torch.ConeBeamProjector.from_conrad_config()

    projection = projector.new_projection_tensor(
        requires_grad=True if with_backward else False)

    projection += 1000.

    result = projector.project_backward(projection, use_texture=with_texture)
    import pyconrad.autoinit
    pyconrad.imshow(result)
    assert result.shape == projector._volume_shape

    assert result is not None
    if with_backward:
        assert projection.requires_grad
        assert result.requires_grad

        loss = result.mean()
        loss.backward()


def test_conrad_forward_backward():
    import pytest
    pytest.importorskip("pyconrad")

    projector = pyronn_torch.ConeBeamProjector.from_conrad_config()
    # import conebeam_projector
    # other_projector = conebeam_projector.CudaProjector()

    volume = projector.new_volume_tensor()

    volume += 1.
    result = projector.project_forward(volume, use_texture=False)
    # import pyconrad.autoinit
    # pyconrad.imshow(result)

    reco = projector.project_backward(result, use_texture=False)
    # import pyconrad.autoinit
    # pyconrad.imshow(reco)

    assert result is not None
    assert reco is not None


def test_register_hook():

    was_executed = False

    def require_nonleaf_grad(v):
        def hook(g):
            nonlocal was_executed
            was_executed = True
            v.grad_nonleaf = g

        v.register_hook(hook)

    projector = pyronn_torch.ConeBeamProjector(
        (128, 128, 128), (2.0, 2.0, 2.0), (-127.5, -127.5, -127.5),
        (2, 480, 620), [1.0, 1.0], (0, 0),
        np.array(
            [[[-3.10e+2, -1.20e+03, 0.00e+00, 1.86e+5],
              [-2.40e+2, 0.00e+00, 1.20e+03, 1.44e+5],
              [-1.00e+00, 0.00e+00, 0.00e+00, 6.00e+2]],
             [[-2.89009888e+2, -1.20522754e+3, -1.02473585e-13, 1.86000000e+5],
              [-2.39963440e+2, -4.18857765e+0, 1.20000000e+3, 1.44000000e+5],
              [-9.99847710e-01, -1.74524058e-2, 0.00000000e+0,
               6.00000000e+2]]]))

    x = projector.new_volume_tensor(requires_grad=True)
    require_nonleaf_grad(x)

    loss = projector.project_forward(x)
    loss.mean().backward()

    x.grad_nonleaf
    assert was_executed



#
# Copyright © 2020 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.

"""

"""


import argparse
import tempfile
from glob import glob
from os import makedirs
from os.path import basename, dirname, join
from shutil import copyfile, copytree, rmtree

try:
    import pystencils
    from pystencils.cpu.cpujit import get_cache_config
    from pystencils.data_types import TypedSymbol, create_type
    from pystencils.kernelparameters import FieldPointerSymbol, FieldShapeSymbol
    from pystencils_autodiff.backends.astnodes import TorchModule
    from pystencils_autodiff.framework_integration.astnodes import \
        CustomFunctionCall
    from pystencils_autodiff.framework_integration.printer import \
        FrameworkIntegrationPrinter

    volume = pystencils.fields('volume: float32[3d]')
    projection = pystencils.fields('projection: float32[3d]')
    projection_matrices = pystencils.fields('matrices: float32[3d]')
    inv_matrices = pystencils.fields('inv_matrices: float32[3d]')
    source_points = pystencils.fields('source_points: float32[1d]')
    volume_slice = pystencils.fields('volume_slice: float32[2d]')
    projections_1d = pystencils.fields('projections_1d: float32[2d]')
    ray_vectors = pystencils.fields('ray_vectors: float32[2d]')

    FUNCTIONS = {
            'Cone_Backprojection3D_Kernel_Launcher': CustomFunctionCall('Cone_Backprojection3D_Kernel_Launcher',
                                                 FieldPointerSymbol(projection.name, projection.dtype, const=True),
                                                 FieldPointerSymbol(volume.name, volume.dtype, const=False),
                                                 FieldPointerSymbol(projection_matrices.name,
                                                                    projection_matrices.dtype, const=True),
                                                 FieldShapeSymbol(['matrices'], 0),
                                                 *[FieldShapeSymbol(['volume'], i) for i in range(2, -1, -1)],
                                                 TypedSymbol('volume_spacing_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_z', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_z', create_type('float32'), const=True),
                                                 *[FieldShapeSymbol(['projection'], i) for i in range(2, 0, -1)],
                                                 TypedSymbol('projection_multiplier',
                                                             create_type('float32'), const=True), backend='gpucuda',
                                                 fields_accessed=[volume, projection, projection_matrices], custom_signature="""
void Cone_Backprojection3D_Kernel_Launcher(const float *sinogram_ptr, float *out, const float *projection_matrices, const int number_of_projections,
                                          const int volume_width, const int volume_height, const int volume_depth,
                                          const float volume_spacing_x, const float volume_spacing_y, const float volume_spacing_z,
                                          const float volume_origin_x, const float volume_origin_y, const float volume_origin_z,
                                          const int detector_width, const int detector_height, const float projection_multiplier);
    """),  # noqa
    'Cone_Projection_Kernel_Launcher': CustomFunctionCall('Cone_Projection_Kernel_Launcher',
                                                 FieldPointerSymbol(volume.name, volume.dtype, const=True),
                                                 FieldPointerSymbol(projection.name, projection.dtype, const=False),
                                                 FieldPointerSymbol(inv_matrices.name,
                                                                    inv_matrices.dtype, const=True),
                                                 FieldPointerSymbol(source_points.name,
                                                                    source_points.dtype, const=True),
                                                 FieldShapeSymbol([source_points.name], 0),
                                                 *[FieldShapeSymbol(['volume'], i) for i in range(2, -1, -1)],
                                                 TypedSymbol('volume_spacing_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_z', create_type('float32'), const=True),
                                                 *[FieldShapeSymbol(['projection'], i) for i in range(2, 0, -1)],
                                                 TypedSymbol('step_size', create_type('float32'), const=True),
                                                 backend='gpucuda',
                                                 fields_accessed=[volume, projection, inv_matrices, source_points],
                                                 custom_signature="""
void Cone_Projection_Kernel_Launcher(const float* volume_ptr, float *out, const float *inv_AR_matrix, const float *src_points, 
                                    const int number_of_projections, const int volume_width, const int volume_height, const int volume_depth, 
                                    const float volume_spacing_x, const float volume_spacing_y, const float volume_spacing_z,
                                    const int detector_width, const int detector_height, const float step_size);
    """),  # noqa
    'Cone_Projection_Kernel_Tex_Interp_Launcher': CustomFunctionCall('Cone_Projection_Kernel_Tex_Interp_Launcher',
                                                 FieldPointerSymbol(volume.name, volume.dtype, const=True),
                                                 FieldPointerSymbol(projection.name, projection.dtype, const=False),
                                                 FieldPointerSymbol(inv_matrices.name,
                                                                    inv_matrices.dtype, const=True),
                                                 FieldPointerSymbol(source_points.name,
                                                                    source_points.dtype, const=True),
                                                 FieldShapeSymbol([source_points.name], 0),
                                                 *[FieldShapeSymbol(['volume'], i) for i in range(2, -1, -1)],
                                                 TypedSymbol('volume_spacing_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_z', create_type('float32'), const=True),
                                                 *[FieldShapeSymbol(['projection'], i) for i in range(2, 0, -1)],
                                                 TypedSymbol('step_size', create_type('float32'), const=True),
                                                 backend='gpucuda',
                                                 fields_accessed=[volume, projection, inv_matrices, source_points],
                                                 custom_signature="""
void Cone_Projection_Kernel_Tex_Interp_Launcher(
    const float *volume_ptr, float *out,
    const float *inv_AR_matrix, const float *src_points,
    const int number_of_projections, const int volume_width,
    const int volume_height, const int volume_depth,
    const float volume_spacing_x, const float volume_spacing_y,
    const float volume_spacing_z, const int detector_width,
    const int detector_height, const float step_size);
    """),  # noqa
    'Parallel_Projection2D_Kernel_Launcher': CustomFunctionCall('Parallel_Projection2D_Kernel_Launcher',
                                                 FieldPointerSymbol(volume_slice.name, volume_slice.dtype, const=True),
                                                 FieldPointerSymbol(projections_1d.name,
                                                                    projections_1d.dtype, const=False),
                                                 FieldPointerSymbol(ray_vectors.name,
                                                                    ray_vectors.dtype, const=True),
                                                 FieldShapeSymbol([ray_vectors.name], 0),
                                                 *[FieldShapeSymbol(['volume_slice'], i) for i in range(1, -1, -1)],
                                                 TypedSymbol('volume_spacing_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_y', create_type('float32'), const=True),
                                                 FieldShapeSymbol([projections_1d.name], 1),
                                                 TypedSymbol('detector_spacing', create_type('float32'), const=True),
                                                 TypedSymbol('detector_origin', create_type('float32'), const=True),
                                                 backend='gpucuda',
                                                 fields_accessed=[volume_slice, projections_1d, ray_vectors], custom_signature="""
void Parallel_Projection2D_Kernel_Launcher(
    const float *volume_ptr, float *out, const float *ray_vectors,
    const int number_of_projections, const int volume_width,
    const int volume_height, const float volume_spacing_x,
    const float volume_spacing_y, const float volume_origin_x,
    const float volume_origin_y, const int detector_size,
    const float detector_spacing, const float detector_origin);
    """),  # noqa
    'Parallel_Backprojection2D_Kernel_Launcher': CustomFunctionCall('Parallel_Backprojection2D_Kernel_Launcher',
                                                 FieldPointerSymbol(projections_1d.name,
                                                                    projections_1d.dtype, const=True),
                                                 FieldPointerSymbol(volume_slice.name, volume_slice.dtype, const=False),
                                                 FieldPointerSymbol(ray_vectors.name,
                                                                    ray_vectors.dtype, const=True),
                                                 FieldShapeSymbol([ray_vectors.name], 0),
                                                 *[FieldShapeSymbol(['volume_slice'], i) for i in range(1, -1, -1)],
                                                 TypedSymbol('volume_spacing_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_spacing_y', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_x', create_type('float32'), const=True),
                                                 TypedSymbol('volume_origin_y', create_type('float32'), const=True),
                                                 FieldShapeSymbol([projections_1d.name], 1),
                                                 TypedSymbol('detector_spacing', create_type('float32'), const=True),
                                                 TypedSymbol('detector_origin', create_type('float32'), const=True),
                                                 backend='gpucuda',
                                                 fields_accessed=[volume_slice, projections_1d, ray_vectors], custom_signature="""
void Parallel_Backprojection2D_Kernel_Launcher(const float *sinogram_ptr, float *out, const float *ray_vectors, const int number_of_projections,
                                               const int volume_width, const int volume_height, const float volume_spacing_x, const float volume_spacing_y,
                                               const float volume_origin_x, const float volume_origin_y,
                                               const int detector_size, const float detector_spacing, const float detector_origin);
    """),  # noqa
}
except Exception as e:
    import warnings
    warnings.warn(str(e))
    TorchModule = None


def get_pyronn_cuda_kernels():
    return glob(join(dirname(__file__), 'PYRO-NN-Layers', '*.cu.cc'))


def get_pyronn_include_paths():
    return join(dirname(__file__), 'PYRO-NN-Layers')


def generate_shared_object(output_folder=None,
                           source_files=None,
                           show_code=False,
                           framework_module_class=TorchModule,
                           generate_code_only=False,
                           update_repo_files=False):

    object_cache = get_cache_config()['object_cache']

    module_name = 'pyronn_torch_cpp'

    if not output_folder:
        output_folder = join(dirname(__file__), '..', '..', 'generated_files')

    if not source_files:
        source_files = glob(join(dirname(__file__), 'PYRO-NN-Layers', '*.cu.cc'))

    cuda_sources = []
    makedirs(join(object_cache, module_name), exist_ok=True)
    rmtree(join(object_cache, module_name, 'helper_headers'), ignore_errors=True)
    copytree(join(dirname(__file__), 'PYRO-NN-Layers', 'helper_headers'),
             join(object_cache, module_name, 'helper_headers'))
    if update_repo_files:
        rmtree(join(output_folder, 'helper_headers'), ignore_errors=True)
        copytree(join(dirname(__file__), 'PYRO-NN-Layers', 'helper_headers'),
                 join(output_folder, 'helper_headers'))

    for s in source_files:
        dst = join(object_cache, module_name, basename(s).replace('.cu.cc', '.cu'))
        copyfile(s, dst)  # Torch only accepts *.cu as CUDA
        cuda_sources.append(dst)
        if update_repo_files:
            dst = join(output_folder, basename(s).replace('.cu.cc', '.cu'))
            copyfile(s, dst)  # Torch only accepts *.cu as CUDA

    module = framework_module_class(module_name, FUNCTIONS.values())

    if show_code:
        pystencils.show_code(module, custom_backend=FrameworkIntegrationPrinter())

    if generate_code_only:
        return module

    extension = module.compile(extra_source_files=cuda_sources,
                               extra_cuda_flags=['-arch=sm_35'],
                               with_cuda=True,
                               compile_module_name=module_name)

    shared_object_file = module.compiled_file
    copyfile(shared_object_file, join(output_folder, module_name + '.so'))

    if update_repo_files:
        with open(join(output_folder, 'pyronn_torch.cpp'), 'w') as f:
            f.write(pystencils.get_code_str(module, custom_backend=FrameworkIntegrationPrinter()))

    return extension


def compile_shared_object(output_folder=None, source_files=None):
    """
    Same as generate_shared_object without pystencils dependency
    """
    object_cache = tempfile.gettempdir()
    module_name = 'pyronn_torch_cpp'

    if not output_folder:
        output_folder = dirname(__file__)

    if not source_files:
        source_files = get_pyronn_cuda_kernels()

    generated_file = join(dirname(__file__), 'pyronn_torch.cpp')

    cuda_sources = []
    makedirs(join(object_cache, module_name), exist_ok=True)
    rmtree(join(object_cache, module_name, 'helper_headers'), ignore_errors=True)
    copytree(join(dirname(__file__), 'PYRO-NN-Layers', 'helper_headers'),
             join(object_cache, module_name, 'helper_headers'))

    for s in source_files:
        dst = join(object_cache, module_name, basename(s).replace('.cu.cc', '.cu'))
        copyfile(s, dst)  # Torch only accepts *.cu as CUDA
        cuda_sources.append(dst)

    # extension = module.compile(extra_source_files=cuda_sources,
        # extra_cuda_flags=['-arch=sm_35'],
        # with_cuda=True,
        # compile_module_name=module_name)

    from torch.utils.cpp_extension import load
    torch_extension = load(module_name,
                           [generated_file] + cuda_sources,
                           with_cuda=True,
                           extra_cflags=[],
                           extra_cuda_cflags=['-arch=sm_35'],
                           build_directory=join(object_cache, module_name),
                           extra_include_paths=[])
    shared_object_file = join(object_cache, module_name, module_name + '.so')
    copyfile(shared_object_file, join(output_folder, module_name + '.so'))
    return torch_extension


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('--output-folder',  default=None)
    parser.add_argument('--source-files',  default=None)
    args = parser.parse_args()

    generate_shared_object(args.output_folder, args.source_files, show_code=True, update_repo_files=True)
    # compile_shared_object(args.output_folder, args.source_files)


if __name__ == '__main__':
    main()



# Copyright [2019] [Christopher Syben]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Makes every implemented operator in python available under the namespace pyronn_layers
# PYRO-NN is developed as an Open Source project under the Apache License, Version 2.0.
#
import os.path
import tensorflow as tf
import pyronn_layers


if tf.test.is_built_with_cuda():
    _pyronn_layers_module = tf.load_op_library(os.path.dirname(__file__)+'/pyronn_layers.so')
    ''' TODO: Improve the getattr method to add only real kernel methods and not everything '''
    for obj in dir(_pyronn_layers_module):
        setattr(pyronn_layers, obj, getattr(_pyronn_layers_module, obj))





#
# Copyright © 2020 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.

"""
This tests are unrelated to pyronn_torch itself.
But rather for wrapping PYRO-NN for other frameworks.
"""

import tempfile

import pytest

from pyronn_torch.codegen import generate_shared_object


def test_wrap_walberla():
    import pytest
    pytest.importorskip("walberla_app")

    from walberla_app.kernel_call_nodes import WalberlaModule

    generate_shared_object(tempfile.TemporaryDirectory, None, show_code=True,
                           framework_module_class=WalberlaModule, generate_code_only=True)


@pytest.mark.xfail(reason="allow failure", strict=False)
def test_wrap_tensorflow():
    import pytest
    pytest.importorskip("pystencils_autodiff")

    from pystencils_autodiff.backends.astnodes import TensorflowModule

    generate_shared_object(tempfile.TemporaryDirectory, None, show_code=True,
                           framework_module_class=TensorflowModule, generate_code_only=True)


def test_wrap_torch():
    import pytest
    pytest.importorskip("pystencils_autodiff")

    from pystencils_autodiff.backends.astnodes import TorchModule

    generate_shared_object(tempfile.TemporaryDirectory, None, show_code=True,
                           framework_module_class=TorchModule, generate_code_only=True)



#
# Copyright © 2020 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.
"""

"""
import numpy as np
import sympy as sp
import torch

import pyronn_torch


class State:
    def __init__(self,
                 projection_shape,
                 volume_shape,
                 source_points,
                 inverse_matrices,
                 projection_matrices,
                 volume_origin,
                 volume_spacing,
                 projection_multiplier,
                 step_size=1.,
                 with_texture=True):
        self.projection_shape = projection_shape
        self.volume_shape = volume_shape
        self.source_points = source_points
        self.inverse_matrices = inverse_matrices
        self.projection_matrices = projection_matrices
        self.volume_origin = volume_origin
        self.volume_spacing = volume_spacing
        self.projection_multiplier = projection_multiplier
        self.with_texture = with_texture
        self.step_size = step_size


class _ForwardProjection(torch.autograd.Function):
    @staticmethod
    def forward(self, volume, state=None):
        if state is None:
            state = self.state
            return_none = True
        else:
            return_none = False

        volume = volume.float().cuda().contiguous()
        projection = torch.zeros(state.projection_shape,
                                 device='cuda',
                                 requires_grad=volume.requires_grad).float().contiguous()

        assert pyronn_torch.cpp_extension
        if state.with_texture:
            pyronn_torch.cpp_extension.call_Cone_Projection_Kernel_Tex_Interp_Launcher(
                inv_matrices=state.inverse_matrices,
                projection=projection,
                source_points=state.source_points,
                step_size=state.step_size,
                volume=volume,
                volume_spacing_x=state.volume_spacing[0],
                volume_spacing_y=state.volume_spacing[1],
                volume_spacing_z=state.volume_spacing[2])
        else:
            pyronn_torch.cpp_extension.call_Cone_Projection_Kernel_Launcher(
                inv_matrices=state.inverse_matrices,
                projection=projection,
                source_points=state.source_points,
                step_size=state.step_size,
                volume=volume,
                volume_spacing_x=state.volume_spacing[0],
                volume_spacing_y=state.volume_spacing[1],
                volume_spacing_z=state.volume_spacing[2])

        self.state = state
        if return_none:
            return projection, None
        else:
            return projection,

    @staticmethod
    def backward(self, projection_grad, state=None, *args):
        if state is None:
            state = self.state
            return_none = True
        else:
            return_none = False

        projection_grad = projection_grad.float().cuda().contiguous()
        volume_grad = torch.zeros(state.volume_shape,
                                  device='cuda',
                                  requires_grad=projection_grad.requires_grad)

        assert pyronn_torch.cpp_extension
        pyronn_torch.cpp_extension.call_Cone_Backprojection3D_Kernel_Launcher(
            state.projection_matrices, projection_grad,
            state.projection_multiplier, volume_grad, *state.volume_origin,
            *state.volume_spacing)

        self.state = state
        if return_none:
            return volume_grad, None
        else:
            return volume_grad,


class _BackwardProjection(torch.autograd.Function):
    backward = staticmethod(_ForwardProjection.forward)
    forward = staticmethod(_ForwardProjection.backward)


class ConeBeamProjector:
    def __init__(self, volume_shape, volume_spacing, volume_origin,
                 projection_shape, projection_spacing, projection_origin,
                 projection_matrices):
        self._volume_shape = volume_shape
        self._volume_origin = volume_origin
        self._volume_spacing = volume_spacing
        self._projection_shape = projection_shape
        self._projection_matrices_numpy = projection_matrices
        self._projection_spacing = projection_spacing
        self._projection_origin = projection_origin
        self._calc_inverse_matrices()

    @classmethod
    def from_conrad_config(cls):
        import pyconrad.autoinit
        import pyconrad.config
        volume_shape = pyconrad.config.get_reco_shape()
        volume_spacing = pyconrad.config.get_reco_spacing()
        volume_origin = pyconrad.config.get_reco_origin()
        projection_shape = pyconrad.config.get_sino_shape()
        projection_spacing = [
            pyconrad.config.get_geometry().getPixelDimensionX(),
            pyconrad.config.get_geometry().getPixelDimensionY(),
        ]
        projection_origin = [
            pyconrad.config.get_geometry().getDetectorOffsetU(),
            pyconrad.config.get_geometry().getDetectorOffsetV(),
        ]
        projection_matrices = pyconrad.config.get_projection_matrices()

        obj = cls(volume_shape=volume_shape,
                  volume_spacing=volume_spacing,
                  volume_origin=volume_origin,
                  projection_shape=projection_shape,
                  projection_spacing=projection_spacing,
                  projection_origin=projection_origin,
                  projection_matrices=projection_matrices)
        return obj

    def new_volume_tensor(self, requires_grad=False):
        return torch.zeros(self._volume_shape,
                           requires_grad=requires_grad).cuda()

    def new_projection_tensor(self, requires_grad=False):
        return torch.zeros(self._projection_shape,
                           requires_grad=requires_grad).cuda()

    def project_forward(self, volume, step_size=1., use_texture=True):
        return _ForwardProjection().apply(
            volume,
            State(projection_shape=self._projection_shape,
                  volume_shape=self._volume_shape,
                  source_points=self._source_points,
                  inverse_matrices=self._inverse_matrices,
                  projection_matrices=self._projection_matrices,
                  volume_origin=self._volume_origin,
                  volume_spacing=self._volume_spacing,
                  projection_multiplier=self._projection_multiplier,
                  step_size=step_size,
                  with_texture=use_texture))[0]

    def project_backward(self,
                         projection_stack,
                         step_size=1.,
                         use_texture=True):
        return _BackwardProjection().apply(
            projection_stack,
            State(projection_shape=self._projection_shape,
                  volume_shape=self._volume_shape,
                  source_points=self._source_points,
                  inverse_matrices=self._inverse_matrices,
                  projection_matrices=self._projection_matrices,
                  volume_origin=self._volume_origin,
                  volume_spacing=self._volume_spacing,
                  projection_multiplier=self._projection_multiplier,
                  step_size=step_size,
                  with_texture=use_texture))[0]

    def _calc_inverse_matrices(self):
        if self._projection_matrices_numpy is None:
            return
        self._projection_matrices = torch.stack(
            tuple(
                torch.from_numpy(p.astype(np.float32))
                for p in self._projection_matrices_numpy)).cuda().contiguous()

        inv_spacing = np.array([1 / s for s in self._volume_spacing],
                               np.float32)

        camera_centers = list(map(
            lambda x: np.array(sp.Matrix(x).nullspace(), np.float32),
            self._projection_matrices_numpy))

        source_points = list(map(
            lambda x: (x[0, :3, 0] / x[0, 3, 0] * inv_spacing
                       - np.array(list(self._volume_origin)) * inv_spacing).astype(np.float32), camera_centers))

        scaling_matrix = np.array([[inv_spacing[0], 0, 0], [0, inv_spacing[1], 0], [0, 0, inv_spacing[2]]])
        inv_matrices = list(map(
            lambda x:
            (scaling_matrix @ np.linalg.inv(x[:3, :3])).astype(np.float32),
            self._projection_matrices_numpy))

        self._inverse_matrices = torch.stack(
            tuple(map(torch.from_numpy, inv_matrices))).float().cuda().contiguous()
        self._source_points = torch.stack(
            tuple(map(torch.from_numpy, source_points))).float().cuda().contiguous()
        self._projection_multiplier = 1. / self._projection_matrices.shape[0]

    @property
    def projection_matrices(self):
        return self._projection_matrices_numpy

    @projection_matrices.setter
    def projection_matrices(self, numpy_matrices):
        self._projection_matrices_numpy = numpy_matrices
        self._calc_inverse_matrices()



import sys
from glob import glob
from os.path import join

from pkg_resources import VersionConflict, require
from setuptools import setup

from torch.utils.cpp_extension import BuildExtension, CUDAExtension

try:
    require('setuptools>=38.3')
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ == "__main__":

    module_name = 'pyronn_torch_cpp'

    cuda_sources = glob(join('generated_files', '*.cu'))

    generated_file = join('generated_files', 'pyronn_torch.cpp')

    setup(use_pyscaffold=True,
          ext_modules=[
              CUDAExtension(module_name,
                            [generated_file] + cuda_sources,
                            extra_compile_args={'cxx': [],
                                                'nvcc': ['-arch=sm_35', '-O3', '-allow-unsupported-compiler']})
          ],
          cmdclass={
              'build_ext': BuildExtension
          })



# -*- coding: utf-8 -*-

import sys
from os.path import dirname

from pkg_resources import DistributionNotFound, get_distribution

from pyronn_torch.conebeam import ConeBeamProjector

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = 'pyronn-torch'
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = 'unknown'
finally:
    del get_distribution, DistributionNotFound


try:
    sys.path.append(dirname(__file__))
    cpp_extension = __import__('pyronn_torch_cpp')
except Exception as e:
    import warnings
    warnings.warn(str(e))
    import pyronn_torch.codegen
    cpp_extension = pyronn_torch.codegen.compile_shared_object()

__all__ = ['ConeBeamProjector', 'cpp_extension']
